package figure;

import java.util.*;

public class ListaLinkataDoppia<E>
implements java.util.List<E>
{
	int numElem;
	NodoDoppio<E> head, tail;

	public ListaLinkataDoppia()
	{
		numElem = 0;
		head = tail = null;
	}

	public boolean add(E e)
	{
		NodoDoppio<E> tmp = new NodoDoppio<E>();
		tmp.contenuto = e;
		if(isEmpty())
		{
			head = tail = tmp;
		}
		else
		{
			tail.next = tmp;
			tmp.prev = tail;
			tail = tail.next;
		}
		numElem++;
		return true;
	}

	public void add(int index, E e)
	{
		if(index < 0 || index > size())
		{
			throw new IndexOutOfBoundsException("indice non valido <" + index + ">");
		}

		if(index == size())
		{
			add(e);
			return;
		}

		NodoDoppio<E> tmp = new NodoDoppio<E>();
		if(index == 0)
		{
			tmp.contenuto = e;
			tmp.next = head;
			head.prev = tmp;
			head = tmp;
			numElem++;
			return;
		}

		NodoDoppio<E> cursore = head;
		while(index > 0)
		{
			cursore = cursore.next;
			index--;
		}
		tmp.next = cursore;
		tmp.prev = cursore.prev;
		tmp.next.prev = tmp;
		tmp.prev.next = tmp;
		numElem++;
		return;
	}

	public boolean addAll(Collection<? extends E> c)
	{
		Iterator<? extends E> iter = c.iterator();
		while(iter.hasNext())
		{
			E tmp = iter.next();
			add(tmp);
		}
		return true;
	}

	public boolean addAll(int index, Collection<? extends E> c)
	{
		if(index < 0 || index > size())
		{
			throw new IndexOutOfBoundsException("indice non valido <" + index + ">");
		}
		if(isEmpty())
		{
			return addAll(c);
		}

		if(index == size())
		{
			return addAll(c);
		}

		NodoDoppio<E> cursore = head;
		Boolean moveHead = (index == 0);
		while(index > 0)
		{
			cursore = cursore.next;
			index--;
		}
		Iterator<? extends E> iter = c.iterator();
		while(iter.hasNext())
		{
			NodoDoppio<E> tmp = new NodoDoppio<E>();
			tmp.contenuto = iter.next();
			if(moveHead)
			{
				head = tmp;
				moveHead = false;
			}
			tmp.next = cursore;
			tmp.prev = cursore.prev;
			tmp.prev.next = tmp;
			tmp.next.prev = tmp;
			numElem++;
		}
		return true;
	}

	public void clear()
	{
		// se ho la certezza che non esce alcun riferimento ad alcun nodo
		// questo e' sufficiente
		/*
		head = tail = null;
		numElem = 0;
		*/
		//Altrimenti, sgancio tutti i singoli puntatori
		NodoDoppio<E> cursore = head;
		while(cursore != null)
		{
			cursore.contenuto = null;
			if(cursore.prev != null)
			{
				cursore.prev.next = null;
			}
			cursore.prev = null;
			cursore = cursore.next;
		}
		head = tail = null;
		numElem = 0;
	}

	public boolean contains(Object o)
	{
		// l'operatore instanceof necessita di un tipo non parametrico, fissato staticamente
		/*
		if( ! (o instanceof E))
		{
			return false;
		}
		*/

		if(o == null)
		{
			NodoDoppio<E> cursore = head;
			while(cursore != null)
			{
				if(cursore.contenuto == null)
				{
					return true;
				}
				cursore = cursore.next;
			}
		}
		else
		{
			NodoDoppio<E> cursore = head;
			while(cursore != null)
			{
				if(o.equals(cursore.contenuto))
				{
					return true;
				}
				cursore = cursore.next;
			}
		}
		return false;
	}

	public boolean containsAll(Collection<?> c)
	{
		Iterator<?> iter = c.iterator();
		while(iter.hasNext())
		{
			if(!contains(iter.next()))
			{
				return false;
			}
		}
		return true;
	}

	public boolean equals(Object o)
	{
		if(! (o instanceof List))
		{
			return false;
		}

		List l = (List)o;

		if(l.size() != size())
		{
			return false;
		}

		NodoDoppio<E> cursore = head;
		Iterator iter = l.iterator();

		while(iter.hasNext())
		{
			Object tmp = iter.next();
			if(tmp == null)
			{
				if( cursore.contenuto != null)
				{
					return false;
				}
			}
			else
			{
				if(! tmp.equals(cursore.contenuto))
				{
					return false;
				}
			}
			cursore = cursore.next;
		}
		return true;
	}

	public E get(int index)
	{
		if(index < 0 || index > size())
		{
			throw new IndexOutOfBoundsException("indice non valido <" + index + ">");
		}
		NodoDoppio<E> cursore = head;

		while(index-- > 0)
		{
			cursore = cursore.next;
		}
		return cursore.contenuto;
	}

	public int indexOf(Object o)
	{
		// l'operatore instanceof richiede un tipo non parametrico, fissato staticamente
		/*
		if(! (o instanceof E))
		{
			return -1;
		}
		*/

		int retval = 0;
		NodoDoppio<E> cursore = head;
		while(cursore != null)
		{
			if(o == null)
			{
				if(cursore.contenuto == null)
				{
					return retval;
				}
			}
			else
			{
				if(o.equals(cursore.contenuto))
				{
					return retval;
				}
			}
			retval++;
			cursore = cursore.next;
		}
		return -1;
	}

	public boolean isEmpty()
	{
		return numElem == 0;
	}

	public int lastIndexOf(Object o)
	{
		NodoDoppio<E> cursore = tail;
		int retval = size()-1;

		while(cursore != null)
		{
			if(o == null)
			{
				if(cursore.contenuto == null)
				{
					return retval;
				}
			}
			else
			{
				if(o.equals(cursore.contenuto))
				{
					return retval;
				}
			}
			retval--;
			cursore = cursore.prev;
		}
		return -1;
	}

	public E remove(int index)
	{
		if(index<0 || index >= size())
		{
			throw new IndexOutOfBoundsException("indice non valido <" + index + ">");
		}


		// se c'e' un elemento solo devo annullare testa e coda
		if(size() == 1)
		{
			E retval = head.contenuto;
			head = tail = null;
			numElem = 0;
			return retval;
		}

		// se tolgo dalla testa ma c'e' piu' di un elemento sposto solo la testa
		if(index == 0)
		{
			E retval = head.contenuto;
			head = head.next;
			head.prev.next = null;
			head.prev = null;
			numElem--;
			return retval;
		}

		// se tolgo dalla coda ma c'e' piu' di un elemento sposto solo la coda
		if(index == size()-1)
		{
			E retval = tail.contenuto;
			tail = tail.prev;
			tail.next.prev = null;
			tail.next = null;
			numElem--;
			return retval;
		}

		// tolgo da una posizione intermedia con almeno un elemento prima e dopo
		NodoDoppio<E> cursore = head;
		while(index-- > 0)
		{
			cursore = cursore.next;
		}
		cursore.next.prev = cursore.prev;
		cursore.prev.next = cursore.next;
		cursore.next = cursore.prev = null;
		numElem--;
		return cursore.contenuto;
	}

	public boolean remove(Object o)
	{
		//iteriamo su tutta la lista alla ricerca di o
		NodoDoppio<E> cursore = head;
		int pos = 0;
		if(o == null)
		{
			while(cursore != null)
			{
				if(cursore.contenuto == null)
				{
					//Un solo elemento, annullo tutto
					if(size() == 1)
					{
						head = tail = null;
						numElem = 0;
						return true;
					}
					//devo sganciare la testa
					if(pos == 0)
					{
						head = head.next;
						head.prev.next = null;
						head.prev = null;
						numElem--;
						return true;
					}
					//devo sganciare la coda
					if(pos == size()-1)
					{
						tail = tail.prev;
						tail.next.prev = null;
						tail.next = null;
						numElem--;
						return true;
					}

					//sono in posizione intermedia
					cursore.next.prev = cursore.prev;
					cursore.prev.next = cursore.next;
					cursore.prev = cursore.next = null;
					numElem--;
					return true;
				}
				pos++;
				cursore = cursore.next;
			}
		}
		else
		{
			while(cursore != null)
			{
				if(o.equals(cursore.contenuto))
				{
					//Un solo elemento, annullo tutto
					if(size() == 1)
					{
						head = tail = null;
						numElem = 0;
						return true;
					}
					//devo sganciare la testa
					if(pos == 0)
					{
						head = head.next;
						head.prev.next = null;
						head.prev = null;
						numElem--;
						return true;
					}
					//devo sganciare la coda
					if(pos == size()-1)
					{
						tail = tail.prev;
						tail.next.prev = null;
						tail.next = null;
						numElem--;
						return true;
					}

					//sono in posizione intermedia
					cursore.next.prev = cursore.prev;
					cursore.prev.next = cursore.next;
					cursore.prev = cursore.next = null;
					numElem--;
					return true;
				}
				pos++;
				cursore = cursore.next;
			}
		}
		return false;
	}

	public boolean removeAll(Collection<?> c)
	{
		NodoDoppio<E> cursore = head;
		int pos = 0;
		boolean retval = false;
		while(cursore != null)
		{
			if(c.contains(cursore.contenuto))
			{
				//Un solo elemento, annullo tutto
				if(size() == 1)
				{
					head = tail = null;
					numElem = 0;
					return true;
				}
				//devo sganciare la testa
				if(pos == 0)
				{
					head = head.next;
					head.prev.next = null;
					head.prev = null;
					numElem--;
					pos--;
					retval = true;
				}
				else if(pos == size()-1) //devo sganciare la coda
				{
					tail = tail.prev;
					tail.next.prev = null;
					tail.next = null;
					numElem--;
					return true;
				}
				else
				{
					//sono in posizione intermedia
					cursore.next.prev = cursore.prev;
					cursore.prev.next = cursore.next;
					cursore.prev = null;
					numElem--;
					pos--;
					retval = true;
				}
				pos++;
				if(pos == 0)
				{
					cursore = head;
				}
				else
				{
					cursore = cursore.next;
				}
			}
		}
		return retval;
	}

	public boolean retainAll(Collection<?> c)
	{
		NodoDoppio<E> cursore = head;
		int pos = 0;
		boolean retval = false;
		while(cursore != null)
		{
			if(!c.contains(cursore.contenuto))
			{
				//Un solo elemento, annullo tutto
				if(size() == 1)
				{
					head = tail = null;
					numElem = 0;
					return true;
				}
				//devo sganciare la testa
				if(pos == 0)
				{
					head = head.next;
					head.prev.next = null;
					head.prev = null;
					numElem--;
					pos--;
					retval = true;
				}
				else if(pos == size()-1) //devo sganciare la coda
				{
					tail = tail.prev;
					tail.next.prev = null;
					tail.next = null;
					numElem--;
					return true;
				}
				else
				{
					//sono in posizione intermedia
					cursore.next.prev = cursore.prev;
					cursore.prev.next = cursore.next;
					cursore.prev = null;
					numElem--;
					pos--;
					retval = true;
				}
				pos++;
				if(pos == 0)
				{
					cursore = head;
				}
				else
				{
					cursore = cursore.next;
				}
			}
		}
		return retval;
	}

	public E set(int index, E elem)
	{
		if(index < 0 || index >= size())
		{
			throw new IndexOutOfBoundsException("indice non valido <" + index + ">");
		}

		E retval = null;
		NodoDoppio<E> cursore = head;
		while(index-- > 0)
		{
			cursore = cursore.next;
		}
		retval = cursore.contenuto;
		cursore.contenuto = elem;
		return retval;
	}

	public int size()
	{
		return numElem;
	}

	public List<E> subList(int from, int to)
	{
		if(from < 0 || to > size()-1 || from > to)
		{
			throw new IndexOutOfBoundsException("indici non validi <" + from + "> <" + to + ">");
		}
		List<E> retval = new ListaLinkataDoppia<E>();
		NodoDoppio<E> cursore = head;
		to -= from;
		while(from-- > 0)
		{
			cursore = cursore.next;
		}

		while(to-- > 0)
		{
			retval.add(cursore.contenuto);
			cursore = cursore.next;
		}
		return retval;
	}

	public Iterator<E> iterator()
	{
		return new IteratoreListaLinkataDoppia<E>(this);
	}

	public ListIterator<E> listIterator()
	{
		return listIterator(0);
	}

	public ListIterator<E> listIterator(int index)
	{
		throw new UnsupportedOperationException();
	}

	public Object[] toArray()
	{
		Object[] retval = new Object[size()];
		Iterator<E> iter = iterator();
		int i = 0;
		while(iter.hasNext())
		{
			retval[i++] = iter.next();
		}
		return retval;
	}

	public <T> T[] toArray(T[] t)
	{
		if(size() > t.length)
		{
			t = (T[])java.lang.reflect.Array.newInstance(t.getClass().getComponentType(), size());
		}
		Iterator<E> iter = iterator();
		int i = 0;
		while(iter.hasNext())
		{
			try
			{
				t[i++] = (T)iter.next();
			}
			catch(ClassCastException cce)
			{
				ArrayStoreException ase = new ArrayStoreException();
				ase.initCause(cce);
				throw ase;
			}
		}
		return t;
	}

	public String toString()
	{
		String retval = "[";
		Iterator<E> iter = iterator();
		while(iter.hasNext())
		{
			retval = retval.concat(", " + String.valueOf(iter.next()));
		}
		retval = retval.concat("]");
		return retval;
	}

	class IteratoreListaLinkataDoppia<E>
	implements Iterator<E>
	{
		ListaLinkataDoppia<E> lista;
		NodoDoppio<E> cursore;
		int pos;
		int last;

		IteratoreListaLinkataDoppia(ListaLinkataDoppia<E> l)
		{
			lista = l;
			cursore = lista.head;
			pos = 0;
			last = 0;
		}

		public boolean hasNext()
		{
			return (cursore != null);
		}

		public E next()
		{
			if(!hasNext())
			{
				throw new NoSuchElementException();
			}
			last = 1;
			E retval = cursore.contenuto;
			cursore = cursore.next;
			pos++;
			return retval;
		}

		public void remove()
		{
			if(last != 1)
			{
				throw new IllegalStateException();
			}
			last = 0;
			if(size() == 1)
			{
				lista.head = lista.tail = null;
			}
			else if(pos == 1)//stacco la testa
			{
				lista.head.next = null;
				cursore.prev = null;
				lista.head = cursore;
			}
			else if(pos == lista.size())//stacco la coda
			{
				lista.tail = lista.tail.prev;
				lista.tail.next.prev = null;
				lista.tail.next = null;
			}
			else//stacco in mezzo
			{
				NodoDoppio<E> tmp = cursore.prev;
				cursore.prev.prev.next = cursore;
				cursore.prev = cursore.prev.prev;
				tmp.next = tmp.prev = null;
			}
			lista.numElem--;
			pos--;
		}
	}
}
